<?php
namespace APS;

/**
 * Singleton class: Logger
 *
 * @package APS
 */
class Logger 
{	
    const LEVEL_TRACE   = 'TRACE';
    const LEVEL_DEBUG   = 'DEBUG';
    const LEVEL_INFO    = 'INFO';
    const LEVEL_ERROR   = 'ERROR';
    const LEVEL_FATAL   = 'FATAL';

    protected static $_instance;

    /**
     * Prohibited for singleton
     */
    protected function __construct() {}

    public static function get()
	{
        if (self::$_instance) {
            return self::$_instance;
        }
        return self::$_instance = new static();
    }

    public static function init()
    {
        // Initialize Logger instance
        self::get();
	}
	
	public function trace($message)
	{
		$this->_log(self::LEVEL_TRACE, $message);
	}
	
	public function debug($message)
	{
        if (defined("APS_DEVELOPMENT_MODE")) 
            $this->_log(self::LEVEL_DEBUG, $message);
	}
	
	public function info($message)
	{
        $this->_log(self::LEVEL_INFO, $message);
	}
	
	public function error($message)
	{
        $this->_log(self::LEVEL_ERROR, $message);
	}
	
	public function fatal($message)
	{
        $this->_log(self::LEVEL_FATAL, $message);
	}

    protected function _log($level, $message)
    {
    	static $tzSet = false;
    	if (!$tzSet)
    	{
	    	date_default_timezone_set("UTC");
			$tzSet = true;
    	}

        if (defined("APS_DEVELOPMENT_MODE"))
            $message = "[".$this->udate("Y-m-d H:i:s.u T")."] $level " . $message . "\n"; // recording milliseconds in dev mode
        else
            $message = "[".date("Y/m/d h:i:s.u")."] $level " . $message . "\n";

        file_put_contents('php://stderr', $message, FILE_APPEND);
	}

    private function udate($format = 'u', $utimestamp = null) {
        if (is_null($utimestamp))
            $utimestamp = microtime(true);

        $timestamp = floor($utimestamp);
        $milliseconds = round(($utimestamp - $timestamp) * 1000000);

        return date(preg_replace('`(?<!\\\\)u`', $milliseconds, $format), $timestamp);
    }
}
