<?php
namespace APS;

require_once "aps/2/types/typeMapFile.php";

define(__NAMESPACE__ . "\APS2_NS_URI", "http://aps-standard.org/ns/2");

class TypeLibrary
{	
	const XML  = "XML";
	const JSON = "JSON";
	
	const PathIdName = "PathResourceId";
	const LinkedResourceId = "LinkedResourceId";
	
	/*
	 *  URI_ID => array( FILE_NAME => CLASS_NAME),
	 */

	/*
	 *  CLASS_NAME => array( FILE_NAME => URI_ID),
	 */

	private static $typesByTypeNameMap = array(
	);

	private static $schemaByTypeIdMap = array();
	private static $schemaByTypeNameMap = array();
	
	public static function getClassByTypeId($type)
	{
 		if (array_key_exists($type, self::$schemaByTypeIdMap)) {
 			Logger::get()->debug("TypeLibrary: Found registered class '$type'");
 			return self::$schemaByTypeIdMap[$type]->name;
 		} else if (array_key_exists($type, $GLOBALS['typesByTypeIdMap'])) {
			$typeRec = $GLOBALS['typesByTypeIdMap'][$type];
			
			Logger::get()->debug("TypeLibrary: loading class '$type' from ".APS_PHP_RUNTIME_ROOT.$typeRec[0]);
			if (!class_exists($typeRec[1])) {
				require_once APS_PHP_RUNTIME_ROOT.$typeRec[0];
			}
			
			self::registerClass($typeRec[1]);
			
			return $typeRec[1];
		} else {
			$cacheDir = 'typeCache';
			if(file_exists($cacheDir.'/typeMapFile.php')) 
				require_once($cacheDir.'/typeMapFile.php');
			if (!isset($GLOBALS['localTypesByTypeIdMap'])) $GLOBALS['localTypesByTypeIdMap'] = array();
			if (array_key_exists($type, $GLOBALS['localTypesByTypeIdMap'])) {
				Logger::get()->info("TypeLibrary: found class '$type' in local type cache");
				$typeRec = $GLOBALS['localTypesByTypeIdMap'][$type];
				if (!class_exists($typeRec[1]) and file_exists($typeRec[0])) {
					require_once $typeRec[0];
					self::registerClass($typeRec[1]);
					return $typeRec[1];			
				}
			} 
			Logger::get()->info("TypeLibrary: fetching class '$type' from controller");
			$typeArray = explode('#', $type);  // typeId [ , structureId ]
			$typeBase = $typeArray[0];
			
			$representation = Request::getController()->getType($typeBase);
			$typeName = getNamespaceByType($representation->id).'\\'.$representation->name;
			Logger::get()->info($representation->id.";".$representation->name);
			if(defined('APS_RUNTIME_HOME'))
				$cacheDir = rtrim(constant('APS_RUNTIME_HOME'), '\\');
			$pathToSource = preg_replace("/\/|\\\\/", DIRECTORY_SEPARATOR, $cacheDir.$typeName).'.'.
									getTypeVersion($representation->id).'.php';

			if(!file_exists($pathToSource)) 
			{
				$obj  = json_encode($representation);
				$schema = new Schema(array('schema' => $obj));

				if (!file_exists(dirname($pathToSource)))
					mkdir(dirname($pathToSource), 0777, true);

				file_put_contents($pathToSource, $schema->toSource());		
			}
			if(!class_exists($typeName))
			{
				Logger::get()->debug("TypeLibrary: Loading type $typeName from $pathToSource");
				require_once $pathToSource;
			}

			self::registerClass($typeName);
			
			$GLOBALS['localTypesByTypeIdMap'][$type] = array($pathToSource, $typeName);
			self::saveLocalTypeList($cacheDir, $GLOBALS['localTypesByTypeIdMap']);
			if (isset($typeArray[1])) 
				return $typeName = getNamespaceByType($representation->id).'\\'.$typeArray[1];

			return $typeName;			
		}		
	}
	
	public static function saveLocalTypeList($cacheDir, $localTypesByTypeIdMap) {
		$content = '<?php
$GLOBALS[\'localTypesByTypeIdMap\'] = array('."\n";
		foreach ($localTypesByTypeIdMap as $type => $passArray) {
			$typeName = str_replace('\\', '\\\\', $passArray[1]);
			$content .= "\"$type\" => array('".$passArray[0]."', '".$typeName."'),\n";
		}
		$content .= ");\n";
		file_put_contents($cacheDir.'/typeMapFile.php', $content);
	}

	public static function isTypeRegistered($type) {
		return array_key_exists($type, self::$schemaByTypeIdMap);
	}
	
	public static function getSchemaByTypeId($type)
	{		
		if (!array_key_exists($type, self::$schemaByTypeIdMap))
			self::getClassByTypeId($type);

		if (!array_key_exists($type, self::$schemaByTypeIdMap))
			throw new SchemaException("Can't find schema for '$type'");			
			
		return self::$schemaByTypeIdMap[$type];
	}
	
	public static function getSchemaByTypeName($class)
	{
		if (class_exists($class) && !array_key_exists($class, self::$schemaByTypeNameMap))
			self::registerClass($class);
			
		if (!array_key_exists($class, self::$schemaByTypeNameMap))	
			throw new SchemaException("Can't find schema for '$class'");		

		return self::$schemaByTypeNameMap[$class];
	}
	
	public static function getSchemaByType($type)
	{
		if (strpos($type, "://") === False)
			return self::getSchemaByTypeName($type);
		else
			return self::getSchemaByTypeId($type);
	}
	
	public static function getClassByType($type)
	{
		if (strpos($type, "://"))
			return self::getClassByTypeId($type);

		return $type;
	}
	
	
	public static function registerClass($name, $isExtendedMode = true)
	{
		$schema = new Schema();
		$schema->scan($name);

		if (!$schema->id)
			return;
		// Load implemented schemas and merge them
		if ($isExtendedMode && !empty($schema->implements))
			foreach($schema->implements as $t)
				if ($t != $schema->id)
					$schema->merge( self::getSchemaByTypeId($t) );

		self::$schemaByTypeIdMap[$schema->id] = $schema;		
		self::$schemaByTypeNameMap[$schema->name] = $schema;
		foreach($schema->getStructures() as $k => $v)
		{
			$nsArray = explode("\\", $k);
			$shortName = end($nsArray);
			$typeId = $schema->id.'#'.$shortName;
			self::$schemaByTypeIdMap[$typeId] = $v;
			self::$schemaByTypeNameMap[$k] = $v;
		}

		return $schema;
	}

	public static function newResourceByTypeId($type)
	{
		$class = TypeLibrary::getClassByTypeId($type);	
		$resource = new $class();
		$resource->aps = new ResourceMeta($type);
		return $resource;
	}

	public static function newObjectByTypeName($typeName)
	{
		$obj = new $typeName();
		return $obj;
	}

	static function parseResourceHeaderXML($resource, $root)
	{
		$resource->aps->type = $root->getAttributeNodeNS ( APS2_NS_URI, 'type' )->value;
		$resource->aps->id = $root->getAttributeNodeNS ( APS2_NS_URI, 'id' )->value;
		$resource->aps->elementNS = $root->getAttribute ( 'xmlns' );
		$resource->aps->elementName = $root->tagName;
	}

	static function parseResourceXML($resource, $root)
	{	
		self::parseResourceHeaderXML($resource, $root);
		unserializeObject($root, $resource);
		
		$resource->aps->links = Link::loadLinksXML($root->ownerDocument);	
		foreach ($resource->aps->links as $link)
		{
			$name = $link->name;
			$resource->$name = $link->isCollection() ? 
							new CollectionProxy($resource, $link) : new ResourceProxy($link);
		} 		
		return $resource;
	}
	
	static public function newResourceXML($root)
	{
		$type = $root->getAttributeNodeNS ( APS2_NS_URI, 'type' )->value;
		$resource = self::newResourceByTypeId($type);
		return self::parseResourceXML($resource, $root);		
	}

	static function parseResourceJSON($resource, $tree)
	{
		if (!($resource instanceof ResourceBase))
			throw new SchemaException("Invalid \$resource parameter");
		
		if (!$tree || !isset($tree->aps))
			throw new SchemaException("No APS resource in JSON tree");
			
		if (!isset($tree->aps->type))
			throw new SchemaException("No type defined for resource");
		$resource->aps = new ResourceMeta($tree->aps->type);
		if(property_exists($tree->aps, 'id'))
			$resource->aps->id = $tree->aps->id;
	        if (property_exists($tree->aps, 'status'))
                   $resource->aps->status = $tree->aps->status;	
	
		if(isset($tree->aps->x509) && isset($tree->aps->x509->self) && isset($tree->aps->x509->controller)) {
			file_put_contents(self::getConfFile().'.pem', $tree->aps->x509->self);
			file_put_contents(self::getConfFile(), json_encode(array(
								 'appCertificate' => $tree->aps->x509->self,
								 'controllerURI' => $_SERVER['HTTP_APS_CONTROLLER_URI'],
								 'controllerCertificate' => $tree->aps->x509->controller
								 )));
		}
		$schema = self::getSchemaByTypeId($resource->aps->type);

		$resource->aps->links = array();
		$resource->aps->elementNS = NULL;
		$resource->aps->elementName = strtolower($schema->name);
		
		$props = $schema->getProperties();
		
		foreach($tree as $k => $v)
		{		
			if($k == 'aps')
				continue;

			if (isset($v->aps) && isset($v->aps->link)) {
				$link = new Link($v, $k);
				
				if($link->isCollection()) {
					$resource->$k = new CollectionProxy($resource->id(), $link);
				} else {
					$resource->$k = new ResourceProxy($link);
				}

				$resource->aps->links[] = $link;
				continue;
			}
			
			// classify down by tree
			if (isset($props[$k]))
				$resource->$k = unserializeObjectJSON($v, $props[$k]);
		}
		// Check if we have all links defined. 
		// It is needed when resources come as a list, in this case there are no links in resource json.
		$schemas[] = $schema;
		foreach($schema->implements as $implemented_type) {
			$sh = TypeLibrary::getSchemaByTypeId($implemented_type);
			if ($sh) $schemas[] = $sh;
		}
		foreach($schemas as $sh)
		{
			foreach($sh->relations as $k => $v)
			{
				if (property_exists($resource, $k) && $resource->$k) continue;
				$link = new Link();
				$link->name = $k;
				$link->href = Proto::resourcePath($resource->aps->id)."/".$k;

				if ($v->collection) {
					$link->collection = true;
					$resource->$k = new CollectionProxy($resource->id(), $link);
				} else {
					if ($v->required) $link->link = 'strong';
					$resource->$k = new ResourceProxy($link);
				}
				$resource->aps->links[] = $link;
			}
			foreach($sh->properties as $k => $v)
			{
				if (property_exists($resource, $k)) continue;
				$resource->$k = null;
			}
			if ($sh->name == '\org\standard\aps\types\core\resource\Resource') continue;
			foreach($sh->structures as $k => $v)
			{
				if (property_exists($resource, $k) or preg_match('#^\\\#', $k)) continue;
				$resource->$k = TypeLibrary::getStructureFromSchema($v);
			}
		}

		return $resource;
	}

	public static function getStructureFromSchema($sh) {
		$struct = new $sh->name;
		if (property_exists($sh, 'properties')) {
			foreach($sh->properties as $k => $v)
			{
				$struct->$k = null;
			}
		}
		if (property_exists($sh, 'structures')) {
			foreach($sh->structures as $k => $v)
			{
				$struct->$k = TypeLibrary::getSchemaStructure($v);
			}
		}
		return $struct;
	}
	
	public static function removeConfFile() {
		@unlink(self::getConfFile());
		@unlink(self::getConfFile().'.pem');
		@unlink(self::getConfFile().'.apsc.pem');
	}

	public static function getAPSCCertFile() {
		$apscCertFile = realpath(TypeLibrary::getConfFile()).".apsc.pem";
		if (file_exists($apscCertFile)) {
			return $apscCertFile;
		}
		$apsCert = json_decode(file_get_contents(TypeLibrary::getConfFile()))->controllerCertificate;
		file_put_contents($apscCertFile, $apsCert);
		return $apscCertFile;
	}

	public static function getConfFile() {
		if(!isset($_SERVER["HTTP_APS_INSTANCE_ID"])) {
			throw new \Exception("APS-Instance-Id header not set.");
		}
		
		if(defined('APS_RUNTIME_HOME')) {
			$confDir = constant('APS_RUNTIME_HOME');
		} else {
			$confDir = class_exists('\APS\Loader') ? Loader::getHome() : ".";
		}
		
		$confDir .= DIRECTORY_SEPARATOR.'config'.DIRECTORY_SEPARATOR;
		
		if(!file_exists($confDir) || !is_dir($confDir)) {
			mkdir($confDir, 0777, true);
		}
		
		return $confDir.$_SERVER["HTTP_APS_INSTANCE_ID"];
	}
	
	static public function newResourceJSON($tree)
	{
		$type = $tree->aps->type;	
		$resource = self::newResourceByTypeId($type);
		
		return TypeLibrary::parseResourceJSON($resource, $tree);	
	}
	
	static public function parseResource($text)
	{
		$firstChar = substr($text, 0, 1);
		if ($firstChar == '{' || $firstChar == '[')
		{
			// JSON
			$tree = json_decode($text);
			return self::newResourceJSON($tree);
		}
		
		if ($firstChar == '<')
		{
			// XML - default case
			$dom = new \DOMDocument();
			$dom->loadXML($text);
	
			$root = $dom->documentElement;

			return self::newResourceXML($root);
		}
		
		throw new SchemaException("Unknown format of data \"$text\"");
	}


	static  function fillResource($resource, $text)
	{
		
		if (substr($text, 0, 1) == '{')
		{
			// JSON
			$tree = json_decode($text);
			return TypeLibrary::parseResourceJSON($resource, $tree);
		} 
		elseif (substr($text, 0, 1) == '<')
		{
		
			// XML 
			$dom = new \DOMDocument();
			$dom->loadXML($text);
	
			$root = $dom->documentElement;
		}
		else
			throw new \Exception("Invalid resource representation [$text]");
			
		return self::parseResourceXML($resource, $root);
	}
	
	/**
	 * Generate Rest routes map
	 * @param array $classSchemas - array of schema classes, if not specified all classes used
	 */
	static public function makeRestRoutes($classSchemas = NULL)
	{
		if (!$classSchemas)
			$classSchemas = self::$schemaByTypeIdMap;
	
		$map = array( 
			"GET" => array(),
			"PUT" => array(), 
			"POST" => array(), 
			"DELETE" => array()
			);
			
		foreach($classSchemas as $id => $schema)
		{
			if (strpos($id, "#"))
				continue; // Structure

			$class = $schema->name;
			$service = $schema->service;
			
			$map['GET'][$service . '/$schema'] = array(
				"class_name" => $class,
				"method_name" => '_getSchema',
				"arguments" => array(),
				"defaults" => array(),
				"metadata" => array( 
									'url' => 'GET /$schema',										
									'description' => ''
					 			   ),
				"method_flag" => 0
			);
			
			$map['GET'][$service . '/$default'] = array(
				"class_name" => $class,
				"method_name" => '_getDefault',
				"arguments" => array(),
				"defaults" => array(),
				"metadata" => array( 
									'url' => 'GET /$default',										
									'description' => ''
					 			   ),
				"method_flag" => 0
			);
			
			foreach($schema->relations as $relation) {
				$linkPath = $service."/{".TypeLibrary::PathIdName."}/".$relation->name;
				$unlinkPath = $service."/{".TypeLibrary::PathIdName."}/".$relation->name."/{".TypeLibrary::LinkedResourceId."}";

				$linkArgs = array();
				$unlinkArgs = array();
			
				$linkArgs[TypeLibrary::PathIdName] = $unlinkArgs[TypeLibrary::PathIdName] = 1;

				$linkArgs['request_data'] = array();
				$linkArgs['request_data']['type'] = $relation->type;
				
				$unlinkArgs[TypeLibrary::LinkedResourceId] = 1;
	
				$linkCall = array(
					"class_name" => $class,
					"method_name" => $relation->name."Link",
					"arguments" => $linkArgs,
					"defaults" => array(),
					"metadata" => array( 
										'url' => "POST"." ".$relation->name."Link",
										'description' => ''
						 			   ),
					"method_flag" => 0
				);

				$unlinkCall = array(
						"class_name" => $class,
						"method_name" => $relation->name."Unlink",
						"arguments" => $unlinkArgs,
						"defaults" => array(),
						"metadata" => array(
								'url' => "DELETE"." ".$relation->name."Unlink",
								'description' => ''
						),
						"method_flag" => 0
				);
					
				$map['POST'][$linkPath] = $linkCall;
				$map['DELETE'][$unlinkPath] = $unlinkCall;
				
			}
			
			foreach($schema->operations as $o)
			{
				$path = $service;

				if (!$o->static)
					$path .= "/{".TypeLibrary::PathIdName."}";
				
				if ($o->path != '/')
					$path .= $o->path;

				$args = array();
				$defs = array();
				
				if (!$o->static)
					$args[TypeLibrary::PathIdName] = 1;
				
				foreach($o->parameters as $n => $p)
				{
					if($p->kind == 'body')
					{
						$name = 'request_data';
						$args[$name] = array();
						$args[$name]['type'] = $p->type;
					} else 
						$args[$p->name] = TRUE;  
					if (isset($p->default))
						$defs[$name] = $p->default;					
				}
	
				$call = array(
					"class_name" => $class,
					"method_name" => $o->name,
					"arguments" => $args,
					"defaults" => $defs,
					"metadata" => array( 
										'url' => $o->verb." ".$o->path,
										'description' => ''
						 			   ),
					"method_flag" => 0
				);

				if (isset($o->response))
					$call['response'] = $o->response;
					
				$map[$o->verb][$path] = $call;
			}
		}
		return $map;
	}
	
}

/**
 * 
 * Unserialize XML node
 * @param DOMElement $node
 * @param mixed $obj - structure to put result to (either object or array)
 * @param mixed $propertySchema - schema of element being parced (either Schema() or Type())
 * @throws SchemaException
 * @throws Exception
 */
function unserializeObject($node, $obj, $propertySchema = NULL)
{
		if ($node->nodeType == \XML_ELEMENT_NODE) {	
			$name = $node->tagName;

			$children = array();
			foreach($node->childNodes as $n) 
			{
					
				if ($n->nodeType == \XML_ELEMENT_NODE)
					if (!$n->getAttributeNodeNS ( APS2_NS_URI, 'link' ))
						$children[] = $n;
			}
			
			if (!$children)
				return $node->textContent;
	
			if ($propertySchema)
			{
				if ($propertySchema->isArray() && is_array($obj))
				{
					$ret = array();
					$obj = TypeLibrary::newObjectByTypeName($propertySchema->getType());
				}
				else
				{
					$obj = $ret = TypeLibrary::newObjectByTypeName($propertySchema->getType());
				}
			}
			else 
			{
				$ret = $obj;
			}
			$schema = TypeLibrary::getSchemaByTypeName(get_class($obj));

			foreach($schema->getProperties() as $k => $prop)
					$ret->$k = NULL;	
			
			foreach($children as $n) 
			{
				$k = $n->tagName;
				try 
				{
					$p = $schema->getProperty($k);
					$v = unserializeObject($n, $ret, $p);
					
			 		if ($p->isSimple())
			 			$v = typeCastSimple($p->getType(), $v);
 		
					if ($p->isArray())
					{
						if (!is_array($ret->$k)) 
							$ret->$k = array();
						array_push($ret->$k, $v);
					} 
					else
						$ret->$k = $v;
				}
				catch (\Exception $ex)
				{
					echo "WARNING: Can't parse $k:".$ex."\n";
					throw $ex;
				}
					 
			}
						
			return $ret;
		}
			
		throw new \Exception("Unknown node type ".$node->nodeType);
}

function serializeObjectXML($node, $obj)
{
	// Skip empty properties
	if(is_null($obj))
		return;

	if(is_array($obj) && empty($obj))
		return;
		
	$type = get_class( is_array($obj) ? $obj[0] : $obj ); 
	$schema = TypeLibrary::getSchemaByTypeName($type);

	foreach ($schema->getProperties() as $prop)
	{
		try
		{
			$k = $prop->name;

			if (($prop instanceof Attribute) && $prop->isLink())
				continue;
				
			if (empty($obj->$k)) // Skip empty properties - probably should not do this - unset scenario
				continue;

			if ($prop->isSimple()) {
				if (is_array($obj->$k))
					foreach ($obj->$k as $v)
						$node->appendChild( $node->ownerDocument->createElement($prop->name, $v) );
				else
					$node->appendChild( $node->ownerDocument->createElement($prop->name, $obj->$k) );
			}
			else
			{
				if ($prop->isArray())
				{
					foreach($obj->$k as $att)
					{
						$n = $node->ownerDocument->createElement($prop->name);
						serializeObjectXML($n, $att);
						$node->appendChild($n);	
					}
				}
				else
				{
					$n = $node->ownerDocument->createElement($prop->name);
					serializeObjectXML($n, $obj->$k);
					$node->appendChild($n);
				}
			}
		} 
		catch(\Exception $ex)
		{
			// Skip unannotaed parts
		}
	}	
}


function unserializeObjectJSON($obj, $valueSchema)
{
	if (is_null($obj))
		return $obj;

	if ($valueSchema->isArray() && !is_array($obj))
			throw new SchemaException("Serialized object does not matches schema: array vs ".gettype($obj));
	
	if ($valueSchema->isSimple())
		return $obj; // returns either simple value or array of simple values

	// complex type
	$type = $valueSchema->getType();
	$childSchema = TypeLibrary::getSchemaByType($type);
	$type = TypeLibrary::getClassByType($type); // map FQType to native type 
	$props = $childSchema->getProperties();
	
	if ($valueSchema->isArray())
	{
		$ret = array();
		foreach($obj as $o)
		{
			$ret1 = new $type();			
			foreach($props as $k => $p)
				$ret1->$k = unserializeObjectJSON($o->$k, $p);

			$ret[] = $ret1;
		}
	}
	else
	{
		$ret = new $type();
		foreach($props as $k => $p)
			if (isset($obj->$k))
				$ret->$k = unserializeObjectJSON($obj->$k, $p);
	}
		
	return $ret;
}

function unserializeObjectFromJSON($obj, $schema, $objectFromJSON)
{
	$properties = $schema->getProperties();
	
	foreach($properties as $k => $p)
		if (isset($obj->$k))
			if(array_key_exists($p->type, Schema::$simpleTypes))
				$objectFromJSON->$k = $obj->$k;
			else {
				$objectFromJSON->$k = unserializeObjectFromJSON($obj->$k, TypeLibrary::getSchemaByTypeName($p->type), new $p->type);
			}

	return $objectFromJSON;
}

/**
 * Cast string to correct local type
 * @param string $type - simple type name
 * @param string $v    - value to be casted
 * @throws SchemaException
 */
function typeCastSimple($type, $v)
{					
	switch($type) {
		case "integer": 
			$v = (int)$v;					
			break;
		case "float":
			$v = (float)$v;
			break; 
		case "boolean":
			$v = (bool)$v;
			break; 
	}
	return $v;
}

?>
